#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

double *X, *Y, *Z;     /// x,y and z coordinates for each node
long *CX, *CY, *CZ;    /// x,y and z code numbers for each node
long *CB;
long Nn;               /// number of nodes, number of elements
long Ndofn=3;          /// number of degrees of freedom in the node
int Param;             /// parameter for prescribed unit displacement in prescribed direction (x=1,y=2,z=3)
char *Fname;
/*
  Param: 0 ... pure periodic boundary conditions for prescribed unit stresses or strains
         1 ... pure periodic boundary conditions for prescribed unit displacement in x direction
         2 ... pure periodic boundary conditions for prescribed unit displacement in y direction
         3 ... pure periodic boundary conditions for prescribed unit displacement in z direction
         4 ... symmetric boundary conditions with periodic b.c.'s in planes - perpendicular direction is fixed 
         5 ... symmetric boundary conditions with free dofs in planes - perpendicular direction is fixed

*/
int main(int argc,char *argv[])
{
  FILE *in,*out;

  long ii,jj,kk,n,nrn,nrn1,nrn2;
  long codenum;
  double xn,xp,yn,yp,zn,zp;

  fprintf(stdout, "\n\n\n---  GENERATOR OF CODE NUMBERS FOR PERIODIC BOUNDARY CONDITIONS OF 3D RECTANGULAR REPRESENTATIVE VOLUME ELEMENT  ---\n\n\n");
  if (argc < 4){
    fprintf (stderr,"\n Wrong number of command line parameters.");
    fprintf (stderr,"\n Use : %s input_file_name output_file_name prescribed_displacement_direction\n\n", argv[0]);
    return(1);
  }
  in = fopen(argv[1], "r");
  if (in==NULL){
    fprintf (stderr,"\n Input file has not been specified.");
    fprintf (stderr,"\n Try it again!\n\n");
    return(2);
  }

  out = fopen(argv[2], "wt");
  if (out==NULL){
    fprintf (stderr,"\n Output file has not been specified.");
    fprintf (stderr,"\n Try it again!\n\n");
    return(3);
  }
  Fname = argv[2];

  if (sscanf(argv[3], "%d", &Param) != 1)
  {
    fprintf (stderr,"\n Parameter for prescribed unit displacement has not been specified.");
    fprintf (stderr,"\n Try it again!\n\n");
    return(4);
  }

  //input file reading
  //basic information about geometry reading
  fscanf(in, "%lf",&xn);//negative plane x-coordinate
  fscanf(in, "%lf",&xp);//posivite plane x-coordinate
  fscanf(in, "%lf",&yn);//negative plane y-coordinate
  fscanf(in, "%lf",&yp);//posivite plane y-coordinate
  fscanf(in, "%lf",&zn);//negative plane z-coordinate
  fscanf(in, "%lf",&zp);//posivite plane z-coordinate

  //nodes reading
  //number of nodes:
  fscanf(in, "%ld",&Nn);

  //coordinates
  X = new double [Nn];
  Y = new double [Nn];
  Z = new double [Nn];
  
  for (ii = 0; ii < Nn; ii++){
    X[ii] = 0.0;
    Y[ii] = 0.0;
    Z[ii] = 0.0;
  }

  for (ii = 0; ii < Nn; ii++)
    {
      fscanf(in, "%ld",&n);
      fscanf(in, "%lf",&X[ii]);
      fscanf(in, "%lf",&Y[ii]);
      fscanf(in, "%lf",&Z[ii]);

      //round for 5 decimals
      X[ii] = X[ii]*100000.0;
      Y[ii] = Y[ii]*100000.0;
      Z[ii] = Z[ii]*100000.0;
      
      X[ii] = round(X[ii]);
      Y[ii] = round(Y[ii]);
      Z[ii] = round(Z[ii]);
      
      X[ii] = X[ii]/100000.0;
      Y[ii] = Y[ii]/100000.0;
      Z[ii] = Z[ii]/100000.0;
    }

  //code numbers
  CX = new long [Nn];
  CY = new long [Nn];
  CZ = new long [Nn];
  CB = new long [Nn]; //this array is for help

  for (ii = 0; ii < Nn; ii++){
    CX[ii] = 1;
    CY[ii] = 1;
    CZ[ii] = 1;
  }

  //reading and setup of corner nodes:
  fscanf(in, "%ld",&nrn);
  for (ii = 0; ii < nrn; ii++)
    {
      fscanf(in, "%ld",&n);
      CX[n-1] = 0;
      CY[n-1] = 0;
      CZ[n-1] = 0;
    }


  if (Param < 5){
    codenum = 2;
    //reading and setup of nodes on x edges:
    //the first edge
    fscanf(in, "%ld",&nrn); 
    for (ii = 0; ii < Nn; ii++){
      CB[ii] = 0;
    }
    
    for (jj = 0; jj < nrn; jj++){
      fscanf(in, "%ld",&n);
      CB[jj] = n;
      if(CX[n-1] == 1){
	CX[n-1] = codenum;
	codenum++;
      }
      if(CY[n-1] == 1){
	CY[n-1] = codenum;
	codenum++;
      }
      if(CZ[n-1] == 1){
	CZ[n-1] = codenum;
	codenum++;
      }
    }
    
    //other x edges
    for (ii = 0; ii < 3; ii++){
      fscanf(in, "%ld",&nrn);
      for (jj = 0; jj < nrn; jj++){
	fscanf(in, "%ld",&n);
	for(kk = 0; kk < nrn; kk++)
	  if(X[n-1] == X[CB[kk]-1]){
	    if(CX[n-1] == 1)
	      CX[n-1] = CX[CB[kk]-1];
	    if(CY[n-1] == 1)
	      CY[n-1] = CY[CB[kk]-1];
	    if(CZ[n-1] == 1)
	      CZ[n-1] = CZ[CB[kk]-1];
	    break;
	  }
      }
    }
    
    //reading and setup of nodes on y edges:
    //the first edge
    fscanf(in, "%ld",&nrn); 
    for (ii = 0; ii < Nn; ii++){
      CB[ii] = 0;
    }
    
    for (jj = 0; jj < nrn; jj++){
      fscanf(in, "%ld",&n);
      CB[jj] = n;
      if(CX[n-1] == 1){
	CX[n-1] = codenum;
	codenum++;
      }
      if(CY[n-1] == 1){
	CY[n-1] = codenum;
	codenum++;
      }
      if(CZ[n-1] == 1){
	CZ[n-1] = codenum;
	codenum++;
      }
    }
    
    //other y edges
    for (ii = 0; ii < 3; ii++){
      fscanf(in, "%ld",&nrn);
      for (jj = 0; jj < nrn; jj++){
	fscanf(in, "%ld",&n);
	for(kk = 0; kk < nrn; kk++)
	  if(Y[n-1] == Y[CB[kk]-1]){
	    if(CX[n-1] == 1)
	      CX[n-1] = CX[CB[kk]-1];
	    if(CY[n-1] == 1)
	      CY[n-1] = CY[CB[kk]-1];
	    if(CZ[n-1] == 1)
	      CZ[n-1] = CZ[CB[kk]-1];
	    break;
	  }
      }
    }
    
    
    //reading and setup of nodes on z edges:
    //the first edge
    fscanf(in, "%ld",&nrn); 
    for (ii = 0; ii < Nn; ii++){
      CB[ii] = 0;
    }
    
    for (jj = 0; jj < nrn; jj++){
      fscanf(in, "%ld",&n);
      CB[jj] = n;
      if(CX[n-1] == 1){
	CX[n-1] = codenum;
	codenum++;
      }
      if(CY[n-1] == 1){
	CY[n-1] = codenum;
	codenum++;
      }
      if(CZ[n-1] == 1){
	CZ[n-1] = codenum;
	codenum++;
      }
    }
    
    //other z edges
    for (ii = 0; ii < 3; ii++){
      fscanf(in, "%ld",&nrn);
      for (jj = 0; jj < nrn; jj++){
	fscanf(in, "%ld",&n);
	for(kk = 0; kk < nrn; kk++)
	  if(Z[n-1] == Z[CB[kk]-1]){
	    if(CX[n-1] == 1)
	      CX[n-1] = CX[CB[kk]-1];
	    if(CY[n-1] == 1)
	      CY[n-1] = CY[CB[kk]-1];
	    if(CZ[n-1] == 1)
	      CZ[n-1] = CZ[CB[kk]-1];
	    break;
	  }
      }
    }
    
    //reading and setup of nodes on -x plane:
    fscanf(in, "%ld",&nrn1); 
    for (ii = 0; ii < Nn; ii++){
      CB[ii] = 0;
    }
    
    for (jj = 0; jj < nrn1; jj++){
      fscanf(in, "%ld",&n);
      CB[jj] = n;
      if(CX[n-1] == 1){
	CX[n-1] = codenum;
	codenum++;
      }
      if(CY[n-1] == 1){
	CY[n-1] = codenum;
	codenum++;
      }
      if(CZ[n-1] == 1){
	CZ[n-1] = codenum;
	codenum++;
      }
    }
    
    //reading and setup of nodes on +x plane:
    fscanf(in, "%ld",&nrn2);
    for (jj = 0; jj < nrn2; jj++){
      fscanf(in, "%ld",&n);
      for(kk = 0; kk < nrn1; kk++)
	if((Y[n-1] == Y[CB[kk]-1]) && (Z[n-1] == Z[CB[kk]-1])){
	  if(CX[n-1] == 1)
	    CX[n-1] = CX[CB[kk]-1];
	  if(CY[n-1] == 1)
	    CY[n-1] = CY[CB[kk]-1];
	  if(CZ[n-1] == 1)
	    CZ[n-1] = CZ[CB[kk]-1];
	  break;
	}
    }
    
    //reading and setup of nodes on -y plane:
    fscanf(in, "%ld",&nrn1); 
    for (ii = 0; ii < Nn; ii++){
      CB[ii] = 0;
    }
    
    for (jj = 0; jj < nrn1; jj++){
      fscanf(in, "%ld",&n);
      CB[jj] = n;
      if(CX[n-1] == 1){
	CX[n-1] = codenum;
	codenum++;
      }
      if(CY[n-1] == 1){
	CY[n-1] = codenum;
	codenum++;
      }
      if(CZ[n-1] == 1){
	CZ[n-1] = codenum;
	codenum++;
      }
    }
    
    //reading and setup of nodes on +y plane:
    fscanf(in, "%ld",&nrn2);
    for (jj = 0; jj < nrn2; jj++){
      fscanf(in, "%ld",&n);
      for(kk = 0; kk < nrn1; kk++)
	if((X[n-1] == X[CB[kk]-1]) && (Z[n-1] == Z[CB[kk]-1])){
	  if(CX[n-1] == 1)
	    CX[n-1] = CX[CB[kk]-1];
	  if(CY[n-1] == 1)
	    CY[n-1] = CY[CB[kk]-1];
	  if(CZ[n-1] == 1)
	    CZ[n-1] = CZ[CB[kk]-1];
	  break;
	}
    }
    
    //reading and setup of nodes on -z plane:
    fscanf(in, "%ld",&nrn1); 
    for (ii = 0; ii < Nn; ii++){
      CB[ii] = 0;
    }
    
    for (jj = 0; jj < nrn1; jj++){
      fscanf(in, "%ld",&n);
      CB[jj] = n;
      if(CX[n-1] == 1){
	CX[n-1] = codenum;
	codenum++;
      }
      if(CY[n-1] == 1){
	CY[n-1] = codenum;
	codenum++;
      }
      if(CZ[n-1] == 1){
	CZ[n-1] = codenum;
	codenum++;
      }
    }
    
    //reading and setup of nodes on +z plane:
    fscanf(in, "%ld",&nrn2);
    for (jj = 0; jj < nrn2; jj++){
      fscanf(in, "%ld",&n);
      for(kk = 0; kk < nrn1; kk++)
	if((X[n-1] == X[CB[kk]-1]) && (Y[n-1] == Y[CB[kk]-1])){
	  if(CX[n-1] == 1)
	    CX[n-1] = CX[CB[kk]-1];
	  if(CY[n-1] == 1)
	    CY[n-1] = CY[CB[kk]-1];
	  if(CZ[n-1] == 1)
	    CZ[n-1] = CZ[CB[kk]-1];
	  break;
	}
    }
    
    //prescribed displacements:
    //displacement in x direction:
    //nodes on -x plane are fixed in x direction
    //nodes on +x plane are loaded by prescribed unit displacement
    if (Param == 1){
      for (jj = 0; jj < Nn; jj++){
	if (X[jj] == xn)
	  CX[jj] = 0;
	if (X[jj] == xp)
	  CX[jj] = -1;      
      }
    }
    
    //displacement in y direction:
    //nodes on -y plane are fixed in x direction
    //nodes on +y plane are loaded by prescribed unit displacement
    if (Param == 2){
      for (jj = 0; jj < Nn; jj++){
	if (Y[jj] == yn)
	  CY[jj] = 0;
	if (Y[jj] == yp)
	  CY[jj] = -1;      
      }
    }
    
    //displacement in z direction:
    //nodes on -z plane are fixed in x direction
    //nodes on +z plane are loaded by prescribed unit displacement
    if (Param == 3){
      for (jj = 0; jj < Nn; jj++){
	if (Z[jj] == zn)
	  CZ[jj] = 0;
	if (Z[jj] == zp)
	  CZ[jj] = -1;      
      }
    }
  }
  
  //symmetric b.c:
  if ((Param == 4) || (Param == 5)){
    for (jj = 0; jj < Nn; jj++){
      if (X[jj] == xn)
	CX[jj] = 0;
      if (X[jj] == xp)
	CX[jj] = 0;
      if (Y[jj] == yn)
	CY[jj] = 0;
      if (Y[jj] == yp)
	CY[jj] = 0;
      if (Z[jj] == zn)
	CZ[jj] = 0;
      if (Z[jj] == zp)
	CZ[jj] = 0;  
    }
  }

  //code numbers printing:
  fprintf(out,"%ld\n",Nn);
  for(ii=0;ii<Nn;ii++)
    fprintf(out,"%ld  %ld  %ld  %ld\n",ii+1,CX[ii],CY[ii],CZ[ii]);
  
  fclose(in);
  fclose(out);

  fprintf(stdout, "\n Code numbers have been printed into outputfile:  %s\n\n\n", Fname);

  return 0;
}
